"""Reports admin classes."""
from adminsortable2.admin import SortableInlineAdminMixin
from django import forms
from django.contrib import messages
from django.contrib.admin import TabularInline
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.shortcuts import redirect
from django.urls import path, reverse, reverse_lazy
from django.utils.html import format_html, mark_safe
from django.utils.translation import gettext, gettext_lazy as _
from django.views.generic import FormView, TemplateView
from grill.clients import LogProperty
from grpc import RpcError

from ferda.backend import LOGGER, REPORT_MANAGERS, REPORTS
from ferda.constants import LOGGER_SESSION_KEY, LogEntryType, LoggerResult
from ferda.models import Report, ReportParameter
from ferda.models.reports import SyncReportsResult

from .guardian import FerdaGuardedModelAdmin


class BaseReportParameterAdmin(SortableInlineAdminMixin, TabularInline):
    """Base report parameter inline admin."""

    model = ReportParameter
    ordering = ('order', 'name')
    fields = ('name', 'type', 'label')
    readonly_fields = ('name', 'type')

    def has_add_permission(self, request, obj):
        """Report parameter can't be added manually through django admin."""
        return False

    def has_change_permission(self, request, obj):
        """Report parameter does not have its own permissions."""
        return request.user.has_perm('ferda.change_report')

    def has_delete_permission(self, request, obj=None):
        """Report parameter can't be deleted manually through django admin."""
        return False

    def has_view_permission(self, request, obj):
        """Report parameter does not have its own permissions."""
        return request.user.has_perm('ferda.view_report')


class ReportInputParameterAdmin(BaseReportParameterAdmin):
    """Report input parameter inline admin."""

    verbose_name_plural = _('Input report parameters')

    def get_queryset(self, request):
        """Filter input parameters."""
        return super().get_queryset(request).filter(is_input=True)


class ReportOutputParameterAdmin(BaseReportParameterAdmin):
    """Report output parameter inline admin."""

    verbose_name_plural = _('Output report parameters')

    def get_queryset(self, request):
        """Filter output parameters."""
        return super().get_queryset(request).filter(is_input=False)


class ReportAdmin(FerdaGuardedModelAdmin):
    """Report admin."""

    list_display = ('name', 'service', 'label')
    ordering = ('service', 'name',)
    inlines = (ReportInputParameterAdmin, ReportOutputParameterAdmin)
    fields = ('service', 'name', 'label', 'description')
    readonly_fields = ('service', 'name',)

    def get_urls(self):
        """Add custom urls."""
        urls = super().get_urls()
        custom_urls = [
            path(
                'manage_reports/',
                self.admin_site.admin_view(ManageReportsView.as_view()),
                name='ferda_report_manage_reports',
            ),
            path(
                'sync_reports/',
                self.admin_site.admin_view(SyncReportsView.as_view()),
                name='ferda_report_sync_reports',
            ),
        ]
        return custom_urls + urls

    def has_add_permission(self, request):
        """Report cannot be added manually through django admin."""
        return False

    def has_delete_permission(self, request, obj=None):
        """Report cannot be deleted manually through django admin."""
        return False


def format_sync_result(result: SyncReportsResult, title=None):
    """Format sync result."""
    return format_html(
        '''<p>
            <strong>{}</strong><br/>
            {}: <tt>{}</tt><br/>
            {}: <tt>{}</tt><br/>
            {}: <tt>{}</tt><br/>
           </p>''',
        title or gettext('Sync result'),
        gettext('Created'), ', '.join(result.created) or mark_safe('&ndash;'),
        gettext('Deleted'), ', '.join(result.deleted) or mark_safe('&ndash;'),
        gettext('Updated header'), ', '.join(result.updated) or mark_safe('&ndash;'),
    )


class ManageReportsForm(forms.Form):
    """Manage reports form."""

    service = forms.ChoiceField(required=True, choices=((service, service) for service in REPORTS))
    sql = forms.CharField(required=True, widget=forms.Textarea)


class ManageReportsView(PermissionRequiredMixin, FormView):
    """Manage reports view."""

    form_class = ManageReportsForm
    template_name = 'admin/ferda/report/manage_reports.html'
    permission_required = ('ferda.add_report', 'ferda.change_report', 'ferda.delete_report')
    success_url = reverse_lazy('admin:ferda_report_changelist')

    def get_context_data(self, **kwargs):
        """Return view context."""
        context = super().get_context_data(**kwargs)
        context['opts'] = Report._meta
        return context

    def form_valid(self, form):
        """Run SQL in the backend database and display results."""
        context = self.get_context_data(form=form)
        service = form.cleaned_data['service']

        log_entry = LOGGER.create(
            LogEntryType.MANAGE_REPORTS,
            source_ip=self.request.META.get('REMOTE_ADDR', ''),
            session_id=self.request.session[LOGGER_SESSION_KEY],
            properties={'service': LogProperty(service, None)},
            content=form.cleaned_data['sql'],
        )
        with log_entry:
            try:
                REPORT_MANAGERS[service].manage_reports(sql=form.cleaned_data['sql'])
            except RpcError as err:
                log_entry.result = LoggerResult.FAIL
                context['error'] = '{}: {}'.format(err.code(), err.details())
                return self.render_to_response(context)
            else:
                log_entry.result = LoggerResult.SUCCESS

        sync_result = Report.objects.sync_with_list(service=service, reports=REPORTS[service].list_reports())

        messages.success(self.request, format_sync_result(sync_result))
        return super().form_valid(form)


class SyncReportsView(PermissionRequiredMixin, TemplateView):
    """Sync reports view."""

    template_name = 'admin/ferda/report/sync_reports.html'
    permission_required = ('ferda.add_report', 'ferda.change_report', 'ferda.delete_report')

    def get_context_data(self, **kwargs):
        """Return view context."""
        context = super().get_context_data(**kwargs)
        context['opts'] = Report._meta
        return context

    def get(self, request):
        """Show expected sync results."""
        context = self.get_context_data()
        expected_sync_result = {
            service: Report.objects.sync_with_list(service=service, reports=client.list_reports(), dry_run=True)
            for service, client in REPORTS.items()
        }
        context['expected_sync_result'] = expected_sync_result

        return self.render_to_response(context)

    def post(self, request):
        """Perform actual db sync."""
        sync_result = {
            service: Report.objects.sync_with_list(service=service, reports=client.list_reports())
            for service, client in REPORTS.items()
        }
        for service in sync_result:
            messages.success(request, format_sync_result(
                sync_result[service],
                title=_("Sync result of service %(service)s") % {'service': service},
            ))
        return redirect(reverse('admin:ferda_report_changelist'))
