"""Registrar certification REST API."""
from datetime import datetime
from typing import List, Optional

from asgiref.sync import sync_to_async
from django.utils.translation import gettext as _
from grill import Properties
from ninja import Router, Schema
from regal import RegistrarCertification
from regal.exceptions import InvalidData, RegistrarCertificationDoesNotExist, RegistrarDoesNotExist

from ferda.api import Error, InvalidDataError, check_permissions
from ferda.backend import LOGGER, REGISTRAR, REGISTRAR_ADMIN
from ferda.constants import LOGGER_SESSION_KEY, LogEntryType, LoggerResult, LogReferenceType

router = Router(tags=["registry", "registrar"])


async def _check_registrar_of_certification(registrar_id: str, certification_id: str) -> None:
    """Check that certification is linked to registrar.

    Args:
        registrar_id: Registrar identification.
        certification_id: Certification identification.

    Raises:
        RegistrarDoesNotExist on registrar not found
        RegistrarCertificationDoesNotExist on certification not found for a registrar
    """
    certifications = await REGISTRAR.get_registrar_certifications(registrar_id)  # might raise RegistrarDoesNotExist
    if certification_id not in [i.certification_id for i in certifications]:
        raise RegistrarCertificationDoesNotExist  # certification not found (for this registrar)


@router.get(
    "/{registrar_id}/certification/", url_name="registrar_certification",
    response={200: List[RegistrarCertification], 404: Error},
    summary="Get registrar certifications",
)
async def get_registrar_certifications(request, registrar_id: str):
    """Get registrar certifications."""
    await sync_to_async(check_permissions)(request, permissions=["ferda.can_view_registry"])

    log_entry = LOGGER.create(
        LogEntryType.REGISTRAR_CERTIFICATION_VIEW,
        source_ip=request.META.get("REMOTE_ADDR", ""),
        session_id=request.session[LOGGER_SESSION_KEY],
        properties={"registrar_id": registrar_id},
    )
    with log_entry:
        try:
            certifications = await REGISTRAR.get_registrar_certifications(registrar_id)
        except RegistrarDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail=_("Registrar does not exist"))
        log_entry.result = LoggerResult.SUCCESS
        log_entry.references = {LogReferenceType.REGISTRAR: registrar_id}
        if certifications:
            log_entry.properties = {"items_count": str(len(certifications))}

        return certifications


class AddCertificationParams(Schema):
    """Body parameters to 'add_registrar_certification'."""

    classification: int
    file_id: str
    valid_from: Optional[datetime] = None
    valid_to: Optional[datetime] = None


@router.post(
    "/{registrar_id}/certification/", url_name="registrar_certification",
    response={200: None, 400: InvalidDataError, 404: Error},
    summary="Add registrar certification",
)
async def add_registrar_certification(request, registrar_id: str, params: AddCertificationParams):
    """Add registrar certification."""
    await sync_to_async(check_permissions)(request, permissions=["ferda.can_edit_registrar"])

    properties: Properties = {
        "registrar_id": registrar_id,
        "classification": str(params.classification),
        "file_id": params.file_id,
    }
    if params.valid_from:
        properties["valid_from"] = params.valid_from.isoformat()
    if params.valid_to:
        properties["valid_to"] = params.valid_to.isoformat()
    log_entry = LOGGER.create(
        LogEntryType.REGISTRAR_CERTIFICATION_CREATE,
        source_ip=request.META.get("REMOTE_ADDR", ""),
        session_id=request.session[LOGGER_SESSION_KEY],
        properties=properties,
    )
    with log_entry:
        try:
            await REGISTRAR_ADMIN.add_registrar_certification(registrar_id, **params.dict())
        except InvalidData as err:
            log_entry.result = LoggerResult.FAIL
            return 400, InvalidDataError(detail=_("Invalid data"), fields=err.fields)
        except RegistrarDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail=_("Registrar does not exist"))
        log_entry.result = LoggerResult.SUCCESS
        log_entry.references = {LogReferenceType.REGISTRAR: registrar_id}


class UpdateCertificationParams(Schema):
    """Body parameters to 'update_registrar_certification'."""

    classification: Optional[int] = None
    file_id: Optional[str] = None
    valid_to: Optional[datetime] = None


@router.put(
    "/{registrar_id}/certification/{certification_id}/", url_name="registrar_certification",
    response={200: None, 400: InvalidDataError, 404: Error},
    summary="Update registrar certification",
)
async def update_registrar_certification(
    request, registrar_id: str, certification_id: str, params: UpdateCertificationParams
):
    """Update registrar certification."""
    await sync_to_async(check_permissions)(request, permissions=["ferda.can_edit_registrar"])

    properties: Properties = {"registrar_id": registrar_id, "certification_id": certification_id}
    if params.classification:
        properties["set_classification"] = str(params.classification)
    if params.file_id:
        properties["set_file_id"] = params.file_id
    if params.valid_to:
        properties["set_valid_to"] = params.valid_to.isoformat()
    log_entry = LOGGER.create(
        LogEntryType.REGISTRAR_CERTIFICATION_UPDATE,
        source_ip=request.META.get("REMOTE_ADDR", ""),
        session_id=request.session[LOGGER_SESSION_KEY],
        properties=properties,
    )
    with log_entry:
        try:
            await _check_registrar_of_certification(registrar_id, certification_id)
            await REGISTRAR_ADMIN.update_registrar_certification(certification_id, **params.dict())
        except InvalidData as err:
            log_entry.result = LoggerResult.FAIL
            return 400, InvalidDataError(detail=_("Invalid data"), fields=err.fields)
        except RegistrarCertificationDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail=_("Certification does not exist"))
        except RegistrarDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail=_("Registrar does not exist"))
        log_entry.result = LoggerResult.SUCCESS
        log_entry.references = {LogReferenceType.REGISTRAR: registrar_id}


@router.delete(
    "/{registrar_id}/certification/{certification_id}/", url_name="registrar_certification",
    response={200: None, 400: InvalidDataError, 404: Error},
    summary="Delete registrar certification",
)
async def delete_registrar_certification(request, registrar_id: str, certification_id: str):
    """Delete registrar certification."""
    await sync_to_async(check_permissions)(request, permissions=["ferda.can_edit_registrar"])

    log_entry = LOGGER.create(
        LogEntryType.REGISTRAR_CERTIFICATION_DELETE,
        source_ip=request.META.get("REMOTE_ADDR", ""),
        session_id=request.session[LOGGER_SESSION_KEY],
        properties={"registrar_id": registrar_id, "certification_id": certification_id},
    )
    with log_entry:
        try:
            await _check_registrar_of_certification(registrar_id, certification_id)
            await REGISTRAR_ADMIN.delete_registrar_certification(certification_id)
        except InvalidData as err:
            log_entry.result = LoggerResult.FAIL
            return 400, InvalidDataError(detail=_("Invalid data"), fields=err.fields)
        except RegistrarCertificationDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail=_("Certification does not exist"))
        except RegistrarDoesNotExist:
            log_entry.result = LoggerResult.FAIL
            return 404, Error(detail=_("Registrar does not exist"))
        log_entry.result = LoggerResult.SUCCESS
        log_entry.references = {LogReferenceType.REGISTRAR: registrar_id}
